
package diffpatch;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.w3c.dom.Node.*;
import org.w3c.dom.traversal.*;

import java.util.*;

/* The basic idea is as follows:
   1) Each node that is "affected" is modelled as a unique node attached 
      to its own unique parent named "diff"; so for each node that is 
      "affected" actually a sub-tree is maintained and that is what is 
      transmitted to the other end.
   2) Update the <action-list> node for each "affected" node, reflecting the 
      current actions. 
*/

/** Class DiffDomNode is used to model the "affected" node as a DOM node.
 *  It contains all the methods that are necessary to describe the 
 *  <action-list>, attribute-list, Xpath of the "affected" node 
 *  depending on the "action" that occured within the DOM node 
 *  maintained by the application.
 */

public class DiffDomNode {
  NamedNodeMap newNodeAttrib, actionAttrib;

  /* attribList - stores the attributes that are added/modified/deleted
   *              by the application.
   * actionList - stores the actions that are effected by the application.
   * diffNode   - stores the node that represents the "affected" node.
   * diffRoot   - stores the root of the diffNode, named "diff"
   */
  Element attribList, actionList, diffNode, diffRoot;
  Document diffNodeDoc;  //The Document node of the tree rooted at "diffRoot" . 
  private final String ACTION_LIST = "action-list"; 
  
  public DiffDomNode(Document doc, Node child) { 
    diffNodeDoc = doc; 
    diffRoot = (Element)doc.getFirstChild(); 
   

    /* Append the <action-list> node, to the diffRoot node */
    diffNode = doc.createElement(child.getNodeName());
    actionList = doc.createElement(ACTION_LIST);     
    diffNode.appendChild(actionList);
    diffRoot.appendChild(diffNode);
  }  

  /* Assigns the reference to a pre-existing "diff-node", that is maintained 
   * as the value of the hash-table.
   */ 
  public void setDiffNode(Node storedNode) { 
     diffNodeDoc = storedNode.getOwnerDocument();
     diffNode = (Element)storedNode;
     actionList = (Element)storedNode.getFirstChild();
  }
  
  /* Processes the "diff-node" according to the information packed in 
   * the nodeInfo.
   */
  public Node wrapper(DiffNodeInfo nodeInfo) {
    
    /* action - stores the type of action as obtianed from nodeInfo.
     * XPathNames - stores the Xpath expressions for the affected-node
     *              and its related-node (a sibling OR its parent).
     */ 
    String action, XPathNames[];
    boolean removeNewNode = false, isParent = false;  
    Node retNode = null; //The Diff-DOM node to be returned, after processing. 

    if ( nodeInfo.action.startsWith("node") ) {
      if (nodeInfo.action.equals(DiffNodeInfo.ACTION_REMOVE)) {
        /* case 1: If a node in the app-DOM that is created in a session and later
   	           removed, then:
                   a) find a  node in the <action-list>  that has the action 
		      <node-insert>, then return NULL.  
          case 2: If a node from a DOM, that is created from a XML file, is removed then:
		    return (that node);
          For actions like <attr-X>(X=add/delete/modify), <node-insert> and 
	  <node-remove> the isParent = true; as the "affected" node is removed with
	  reference to the parent node.
        */
	
	/* Remove all othe nodes that could have been inserted under <action-list>
	   and add a node <node-remove>
	*/
        removeActionList();  
        addToActionList(nodeInfo.action);
        if (isRecentlyCreated(nodeInfo.storedNode))  {
            retNode = null;
        }
        else  {
           retNode = diffNode;  
       } 
      } //if node-remove
      else if ( nodeInfo.action.equals(DiffNodeInfo.ACTION_INSERT) ) { 
            isParent = true;
            addToActionList(nodeInfo.action);
            retNode = diffNode;
      } 
      else if (nodeInfo.action.equals(DiffNodeInfo.ACTION_INSERTBEFORE)|| 
	       nodeInfo.action.equals(DiffNodeInfo.ACTION_REPLACE) ) {
            isParent = false; 
            boolean isNull = false; 
            /* 
              case 1: A node/sub-tree that's INSERTED could REPLACE another node; =>  
              case 2: A node/sub-tree that's INSERTED-BEFORE  could REPLACE another node; =>
	      If the "affected" node is from a DOM of an XML file, then:
                 append a new <action-list> node. 
              else just update the Xpath attribs, as though the "affected" is inserted there. 
            */
             if (nodeInfo.storedNode == null )  
              /* A node that exists in the appDOM, as read from an existing XML file.
	      */
                 isNull = true;       
             if (isRecentlyCreated(nodeInfo.storedNode) && !isNull) {
	      /* Remove all othe nodes that could have been inserted under <actionList>
    	         and add a node by name <node-insertBefore> OR <node-replace>.
	      */
               removeActionList();
               addToActionList(nodeInfo.action);
                  if (nodeInfo.action.equals(DiffNodeInfo.ACTION_INSERTBEFORE) ) {
                     retNode = null;    
                  }
                  else {
                    retNode = null;  
                   }
               } 
               else  {
                 retNode = diffNode;
               }
      } //else-if 1
      /* Sets the Xpath expressions for the location of the nodes within
       * the Diff-DOM tree.
       */
       setXPathAttribs(nodeInfo.parent, nodeInfo.newChild, true);
    } //"node"
    else if ( nodeInfo.action.startsWith("attr") ) {
       Element attrChild = null;

       // Checks if any node beginnning with "attr" is present in the <actionList>.
       boolean isAttrNodePresent = isAttrActionPresent(nodeInfo.action);  
       setXPathAttribs(nodeInfo.parent, nodeInfo.newChild, true);
       if (nodeInfo.action.equals(DiffNodeInfo.ACTION_ATTRADD)) {
	  /* The first time an attribute is added, insert a <attr-add> node
	     in the "actionList" and thereafter just append the added attribute,
	     value pair to the <att-add> node.
	  */   
          if ( !isAttrNodePresent ) {
            addToActionList(nodeInfo.action);
          }
          attrChild = (Element)actionList.getLastChild();
          attrChild.setAttribute(nodeInfo.attrName, nodeInfo.newValue);     
	  
       }
       else if (nodeInfo.action.equals( DiffNodeInfo.ACTION_ATTRMODIFY)) {
        /* Remove all othe nodes that could have been inserted under <actionList>
           and add a node by name <attr-modify>.
        */
         if ( !isAttrNodePresent ) {
	   removeActionList();
           addToActionList(nodeInfo.action);
         }
         attrChild = (Element)getNodeReference(DiffNodeInfo.ACTION_ATTRMODIFY);
	 /* Remove the attribute whose value has not changed, and update only 
	    those that have changed. 
	  */
	 if (nodeInfo.preValue.equals(nodeInfo.newValue) )
	    attrChild.removeAttribute(nodeInfo.attrName);
	 else if ( !(nodeInfo.preValue.equals(nodeInfo.newValue)) ||
		   nodeInfo.preValue == null )
            attrChild.setAttribute(nodeInfo.attrName, nodeInfo.newValue);  
       } 
       else if (nodeInfo.action.equals(DiffNodeInfo.ACTION_ATTRDELETE) ){
         attrChild.removeAttribute(nodeInfo.attrName);
       }
       retNode = diffNode;
    } //"attr"
 
    return(retNode);
  } //wrapper 

 
    
  /* Sets the Xpath expressions for the location of the nodes within
     the Diff-DOM tree.
  */ 
  public void setXPathAttribs(Node parent, Node newChild, boolean isParent) {
    String XPathNames[] = new String[2];
    System.arraycopy(getXPath(parent, newChild, isParent),0,XPathNames,0,2);
    diffNode.setAttribute("ref", XPathNames[0]);
    diffNode.setAttribute("new", XPathNames[1]);  

  }

  // Checks if any node beginning with <attr> is present in the <actionList>.
  public boolean isAttrActionPresent(String attrAction) {
    boolean flag = false; 
    NodeList tempList = actionList.getChildNodes();
    int len =  tempList.getLength();
    for (int i = 0; i < len; i++) {
      if ( tempList.item(i).getNodeName().equals(attrAction) ) {
        flag = true;
        break; 
      } 
    } 
    return flag;
  }

  // Removes a particular node with the given node-name.
  public void removeActionNode(String nodeName) {
    Node retNode = getNodeReference(nodeName);
    if (retNode != null) {
      actionList.removeChild(retNode);
    }	    
  }

  public void replaceActionNode(String oldname, String newname) {
    Node old = getNodeReference(oldname);
    if (old != null)
       actionList.replaceChild(getNodeReference(newname), getNodeReference(oldname));
    else
       System.out.println("Old..not found");
  }
  
   // Obtians the reference to the node given its name, if it exists.
   public Node getNodeReference(String nodeName) {
     NodeList tempList = actionList.getChildNodes();
     int len =  tempList.getLength();
     Node retNode = null; 
     for (int i = 0; i < len; i++) {
     if ( tempList.item(i).getNodeName().equals(nodeName) ) {
        retNode = tempList.item(i);
        break;  
      }
     }  
    return(retNode);
   }
  
  // Inserts a newly created node, given the name, to the <actionList>.
  public void addToActionList(String action) { 
    Node newnode = diffNodeDoc.createElement(action);
    actionList.appendChild(newnode);  
  }

   /* 
      The function needs to remove all the nodes attached to teh actionList;
      Pitfall:  Try to remove the nodes in a "nodeList" of the "actionList" node; every time
                a node is removed from the List, its length reduces by 1 and hence in the next 
		iteration the index of a non-existent node is removed. (code fails here.)
      How its done: Have a loop such that as long as the length of the list is non-zero,
                    continue to remove the nodes that are at the front (i-e:- with index 0)
       
   */
  
   public void removeActionList() {
    NodeList tempList = actionList.getChildNodes();
    int len =  tempList.getLength();
    while(true) {
      if ( (len = tempList.getLength()) > 0) { 
         actionList.removeChild(tempList.item(0));	     
         tempList = actionList.getChildNodes();
      }
      else
        break; 
    } 
   } 

  // Tells whether the given node is created in the current session(FALSE) OR not(TRUE).
  public boolean isRecentlyCreated(Node newchild) { 
   boolean recentlyCreated = false;
   Node action = newchild.getLastChild(); 
   NodeList actList = action.getChildNodes();
   int i = 0, length = actList.getLength();
   for ( i = 0; i < length; i++) {
     String nodeName = actList.item(i).getNodeName();  
     if ( nodeName.equals("node-insert") || 
          (nodeName.equals("node-insertBefore")) ) {
            recentlyCreated = true;
            break;
      }
   } 
   return (recentlyCreated) ;
  }

 /* Returns the Xpath expression for the target-node and the related-node;
    handles the cases:
    1) A sub-tree could be built but is not yet attached to the root-node
       in which case the Xpath expression starts with "/NULL";
    2) A node is attached to the root-node as soon as it is formed.       
 */

 public String[] getXPath (Node refNode ,Node findNode,  boolean isParent) {
   String[] XPathNames ;
   Node tempNode = refNode;
   XPathNames = new String[2];
    
   if (refNode == null) {
      if (findNode.getNodeType() == Node.ELEMENT_NODE) {
       if ( isParent ) {
         XPathNames[0] =  "NULL/" + findNode.getNodeName();
         XPathNames[1] =  XPathNames[0];
       }
       else {  
         XPathNames[1] = ""; 
         XPathNames[0] = "";
       }
      return (XPathNames);
     }
   }
   else if (refNode.getNodeType() == Node.DOCUMENT_NODE) {
        XPathNames[0] =  "/" ; 
        XPathNames[1] =  XPathNames[0] + findNode.getNodeName();
        return (XPathNames);
   }
  if ( refNode.getNodeType() == Node.ELEMENT_NODE ) {
     System.arraycopy(getXPath(refNode.getParentNode(), refNode, isParent),
                      0, XPathNames, 0, 2);
  }
  
  if (isParent) {
   XPathNames[1] =  XPathNames[1] + "/" + getNodeNameWithPos(refNode, findNode);
   XPathNames[0] = XPathNames[1].substring(0, XPathNames[1].lastIndexOf('/'));
  }
  else {
   /*  Let the first two Xpath names be the same,  until the level of recursion,
       wherein the two nodes relate as siblings and not as parent-child.
   */ 
     if (refNode.getNodeName().equals(findNode.getParentNode().getNodeName()) ) { 
       XPathNames[1] = XPathNames[1] + "/" + findNode.getNodeName();
     }
     else {
       //Extract the portion until the parent of both these nodes are obtained.
       String temp = XPathNames[1].substring(0, XPathNames[1].lastIndexOf('/') + 1);
       XPathNames[0] = temp +  refNode.getNodeName(); 
       XPathNames[1] = temp +  findNode.getNodeName(); 
     }
   }
  return (XPathNames);
  } //getXPath

   public String getNodeNameWithPos(Node related, Node target) { 
    String targetNodePos; 
    int length = related.getChildNodes().getLength();
    int pos = getNodePos(target);
    targetNodePos = target.getNodeName() + "[" + String.valueOf(pos) + "]";
    return (targetNodePos);
  } 

   public int getNodePos(Node target) {
    int pos = 0;	   
    if (target.getPreviousSibling() == null)
       return (pos);
    else { 
       pos = getNodePos(target.getPreviousSibling());
       return(1 + pos);
    }   
   }
  
   /* Updates the Xpath expression of the related-node, that acts as the 
      reference node for operations like "replace", "insertBefore".
   */
   public void updateRefXPathAttrib(Node newnode, Node ref, Hashtable diffTable) {
     String XPathNames[] = new String[2];
     HashValue parentVal, childVal;
     String key = newnode.getNodeName();
     Node storedNode;    
     System.arraycopy(getXPath(newnode, ref, false),0,XPathNames,0,2);
     parentVal = (HashValue)diffTable.get(XPathNames[1]);          
     storedNode = parentVal.domNodeRef;
     ((Element)storedNode).setAttribute("ref", XPathNames[0]);     
  }

  /*  Updates the Xpath expresssion of the target-node for actions, when 
      its position changes when it is "insertedBefore" another node.
      Idea:
      1) Traverse all descendants of this node
      2) For each descendant of this node traverse its list and modify the XPath attributes;
  */
  public void updateXPathAttribs(Node ancestor, Hashtable diffTable) {
   Node diffParent, diffChild, child; 
   HashValue parentVal, childVal;
   String key = ancestor.getNodeName();
   String XPathNames[] = new String[2];
   for (int i=0; i < ancestor.getChildNodes().getLength(); i++) {
     child = ancestor.getChildNodes().item(i);
     System.arraycopy(getXPath(ancestor, child, true), 0 , XPathNames, 0, 2);
     childVal = (HashValue)diffTable.get(XPathNames[1]);
     diffChild  = childVal.domNodeRef;
     System.out.println("child node name is " + child.getNodeName());
     ((Element)diffChild).setAttribute("ref", XPathNames[0]);
     ((Element)diffChild).setAttribute("new", XPathNames[1]);
     updateXPathAttribs(child, diffTable);
   } //for 
 } //updateXPathAttribs

 /* Deletes the children of a node recursively, and reclaims their entries
    in the hash-table also.
    Idea:
    1) Traverse all descendants of this node.
    2) For each descendant of this node, delete its key and its 
       node within the diffDom. 
 */
 public void deleteDiffChildren(Node ancestor, Hashtable diffTable) {
   String XPathNames[] = new String[2];
   Node diffParent, diffChild, child;
   HashValue parentVal, childVal;
   String key;   

   System.out.println("Inside deleteDiffChildren.."); 
   for (int i=0; i < ancestor.getChildNodes().getLength(); i++) { 
     child = ancestor.getChildNodes().item(i); 
     childVal = (HashValue)diffTable.get(XPathNames[1]); 
     deleteDiffChildren(child, diffTable);
   } //for

     System.arraycopy(getXPath(ancestor.getParentNode(), ancestor, true), 0 , 
		      XPathNames, 0, 2);
     childVal = (HashValue)diffTable.get(XPathNames[1]); 
     diffTable.remove(XPathNames[1]);  
 } //deleteDiffChildren()
 
} // Class DiffDomNode 
