
/**
 * Package org.w3c.dom.events is obtained from 
 * apache's Xerces2-J API.
 */ 

package diffpatch;

import org.w3c.dom.events.*;
import org.w3c.dom.events.EventTarget;
import org.w3c.dom.events.MutationEvent.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.w3c.dom.Node.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import java.io.*;
import java.util.*;

import sharedboard.*;
import editor.*;


 /* The basic idea is as follows:
   1) The event listener captures all the events related to attributes,
      node-insertion, node-deletion.
   2) The event is processed by delegating the same to class DiffDomNode.
 */

/**
 * Class DiffListener tracks the events fired when changes are made to
 * the applicn.-DOM  maintained by the application.
 */

public class DiffListener implements org.w3c.dom.events.EventListener {

   /* diffTable - stores the reference to the HashTable object 
    *             that is instantiated in XMLDiff class.
    */

  private Hashtable diffTable;
  
  private Document diffDoc;  // diffDoc - Document instance variable of DiffDOMDoc 
  /* diffChild - stores the information about a newly formed child node 
   *             of the Diff-DOM with the root node as "diff" and 
   *             ways to process the same. 
   */ 
  private DiffDomNode diffChild;

  

   /* diff - creates a Document object for every newly created node of 
    *         the Diff-DOM, with the root nodes as "diff".
    */        
  //DiffDOMDoc diff;
  
  public DiffListener( Hashtable table) { 

   diffTable = table; 
  }

  public void handleEvent(org.w3c.dom.events.Event evt) {
	  
   /* related - the related node of the node that fired the event .    
    * target -  the node that fired the event.
    * diffNode - the node that is stored in the DOM (Diff-DOM) after 
    *            being processed.
    */
    
    Node related, target, diffNode;

    /* key - the key that is used to lookup the HashTable.
     * action - user-defined action name, after trapping the event that 
     *          was fired.
     * preValue, newValue - the previous and new values of the attribute that was affected.
     */
    
    String key, action = null, attrName = null, preValue = null, newValue = null;

    MutationEvent mevt = (MutationEvent) evt; //added 4/24
    target = (Node) (mevt.getTarget());

    if (  mevt.getType().equals("DOMAttrModified") )
      related = target.getParentNode();
    else
      related = mevt.getRelatedNode();  
    
    /* diff - stores a new DiffDOMDoc object whenever the event is fired
     * diffDoc - Document instance variable of DiffDOMDoc.
     * diffChild - the individual node that occurs in the DOM modelling the "diff".
     */
    
    DiffDOMDoc diff = new DiffDOMDoc(); 
    diffDoc = diff.getNewDiffDoc();  
    diffChild = new DiffDomNode(diffDoc, target); 
    
    /* key - stores the key for the target node that has fired the event
       recently.
    */
  
    if (related.getNodeName().equals(
			            target.getParentNode().getNodeName()
			    ))
       key = this.getHashKey(target, related, true);
    else
       key = this.getHashKey(target, related, false);   
    HashValue hashEntry = new HashValue();
    
    /* Form a key with the target's name in a XPath format
       to check for the existence of the key, use this.
       Have to do a substring-matching to see an existing key, that
       has hashed a DOM node - is updated in the appDOM
       i-e: /child/child[1] -> /root/child/child[1]
 
    */ 

    System.out.println("looking for key:: " + key);
    
    /* 
      key DOES exist for operations - removeChild.
      key MAY exist for operations - attr-insert/delete/modify; insb4, 
                                     replace. 
      key DOES NOT exist for appendChild.  
      Irrespective of what happens in applicn.-DOM, Diff-DOM logs the action by 
      creating a node for it.
    */ 

    /* 
      Checking for the nature of the attribute modification (OR) node insertion (OR) 
      node removal event. 
     */
    
     if ( mevt.getType().equals("DOMAttrModified") ) {	
        attrName = mevt.getAttrName();
        preValue = mevt.getPrevValue();
        newValue = mevt.getNewValue();

        switch ( mevt.getAttrChange() ) {
         case MutationEvent.ADDITION: {
            action = DiffNodeInfo.ACTION_ATTRADD;
            break;
         } 
         case MutationEvent.REMOVAL:  {
            action =  DiffNodeInfo.ACTION_ATTRDELETE;
            break;
         }
         case MutationEvent.MODIFICATION: { 
            action =  DiffNodeInfo.ACTION_ATTRMODIFY;
            break;
         }
         default: {
           System.out.println(" DOMAttrModified: unknown attribute event-type" );
        } 
      }
    } 
    else if ( mevt.getType().equals("DOMNodeInserted") ) {
      action = DiffNodeInfo.ACTION_INSERT; 
    } 
    else  
      action = DiffNodeInfo.ACTION_REMOVE; 

     //nodeInfo - stores the context information gathered above. 
    DiffNodeInfo nodeInfo = new DiffNodeInfo(action, related, target,   
  		                             null, attrName, preValue,newValue); 
 
    /* Check if the node ia an "Element" node */
    if (target.getNodeType() == Node.ELEMENT_NODE) {
	    
       /* If this is the first time that the target node has 
	  fired the event, then the hash-key is not yet there.
	*/
	    
       if (  !diffTable.containsKey(key) ) {
          //System.out.println("KEY= " +  key + " not found.."); 
          /* 
	   * diffNode - stores the node that will be inserted into the 
	   *             Diff-DOM, that appropriately stores the "diff"
	   *		information.
	   */
	  diffNode = diffChild.wrapper(nodeInfo);
          if ( action.startsWith("attr") || action.equals(DiffNodeInfo.ACTION_INSERT) ) {
            
            /* The hash-value is the individual "diff" node in the Diff-DOM that bears the 
	       information for the target node, that fired the event, which will be sent 
	       across to the peer.
	    */
	    hashEntry.domNodeRef = diffNode;
            diffTable.put(key, hashEntry);
	    
        } //if(2)                     
      } //if(1)
      else  {
 	   Node storedNode = ((HashValue) diffTable.get(key)).domNodeRef;
	   nodeInfo.storedNode = storedNode;
	   /* 
	      Since the key is found, a reference to the node stored as
	      a part of Diff-DOM exists and that node should be processed further. 
	   */

           diffChild.setDiffNode(storedNode);
	   diffDoc = storedNode.getOwnerDocument(); 
	   diffNode = diffChild.wrapper(nodeInfo); 
	   
           if (action.equals("node-remove") ) {
             /* case 1: A node that was created by the user in a session is
	                removed then wrapper() returns NULL.
                case 2: A node that was already created from a XML file is removed, 
		        then wrapper() returns that node.
             */
                if ( diffNode == null ) { 
                /* The entries in the HashTable are reclaimed by deleting the keys 
		    associated with the target node and its descendants 
		*/
                diffChild.deleteDiffChildren(target, diffTable);  
               }
               else {
	          System.out.println("The node that was created earlier"+ 
				     "from an XML file is deleted.");
               }
           } //if
    } //else
  } //ELEMENT_NODE;
 } // handleEvent()

  /* Returns the Document object with the recently formed Diff-DOM node
     so that it can be serialized and sent across to the peer 
   */
  
 public Document getAffectedDoc() {
   return(diffDoc);  
 } 

 
 public String getHashKey(Node target, Node related, boolean isParent) {
  String  xpath[] = new String[2];
  System.arraycopy(diffChild.getXPath(related, target, isParent), 0, xpath, 0, 2);
  return(xpath[1]);
 }

 //Debugging and testing
 public void printHashKeys() {
  Enumeration e;
  for(e  = diffTable.keys(); e.hasMoreElements(); ) {
     System.out.println("Key = " + e.nextElement() );
  }  
 }

} //DiffListener


