
/**
 * All classes used to do the "difference-generatiion" and 
 * "change-resolver" fall within this package.
 */

package diffpatch;

import org.w3c.dom.events.*;
import org.w3c.dom.events.EventTarget;
import org.w3c.dom.events.MutationEvent.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.w3c.dom.Node.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import java.io.*;
import java.util.*;

import sharedboard.*;  //Importing application-specific classes
import editor.*;

 /* The basic idea is :
    Event listener is attached to the "Document" node of the app-DOM, and there-
    after whatever event happens within its sub-tree is propagated to it. 
 */

/* 
Functions:

 Importance of Hash-table, data struct:
 2 cases:
 1) A sub-tree could be constructed, such that it might replace/insert-before an exisiting node
    If the parent of this sub-tree is not yet attached to a node in the DOM, then
    its appended say; no reference to the child-nodes of this node is there (OR) we don't know that its
    a sub-tree;
    => DS enables to know that key is not present and hence the need to create "diffDOM"
       for each of these nodes;
 2) Place holder for info. like "Xpath", action that "affected" that node.

*/

/**
 * Class "XMLDiff" is the entry-point for the "difference-generation"(diff)
 * module; it creates custom-data structures, registers the event-listener
 * for the "Document" object handed down by the application.
 */

public class  XMLDiff
{
  /* appDoc -  stores a reference to the Document object handed down by 
   *           the appication.
   * diffDoc - Document instance variable of DiffDOMDoc.            
   */ 
  private Document appDoc, diffDoc; 
  
  /* The hashtable used as a custom data-structure, key being the 
   * Xpath formatted expression of the target-node.
   */ 
  private Hashtable table; 
 
  //The event-listener for "difference-generation".
  private DiffListener diffListener; 
  
  //A simple WhiteBoard Application client for talking to the peer's server.
  private WBAppClient wbClient;
  private boolean isInsideEventistener = true; 
  
  public XMLDiff(Document appDOMDoc, String server, 
                 int appServPort, ShapeMgr shapeMgr) {
	  
    appDoc = appDOMDoc;
    table = new Hashtable();  
    
    //Initialize the client connetion etc.  
    wbClient = new WBAppClient(server, appServPort);     
    diffListener = new DiffListener(table);
    addDiffListener();
  }

   /** Registration of the application's Document object with the 
    *  MutationEvent listeners.
    */ 
   public void addDiffListener() { 
    ((EventTarget) appDoc).addEventListener("DOMNodeInserted", diffListener, false);
    ((EventTarget) appDoc).addEventListener("DOMNodeRemoved", diffListener, false);
    ((EventTarget) appDoc).addEventListener("DOMAttrModified", diffListener, false); 
   }

   //Removing the MutationEvent listeners added earlier.
   public void removeDiffListener( ) {
    ((EventTarget) appDoc).removeEventListener("DOMNodeInserted", diffListener, false);
    ((EventTarget) appDoc).removeEventListener("DOMNodeRemoved", diffListener, false);
    ((EventTarget) appDoc).removeEventListener("DOMAttrModified", diffListener, false); 
   }

   public DiffListener getDiffListener() {
     return (diffListener);	   
   }
   /** Sending the Document node of the Diff-DOM serialized to bytes, to
    *  the peer's server.
    */  
   public void sendToWBServer() { 
     wbClient.sendWBMessage(diffListener.getAffectedDoc());  
   }

   /** Wrapper for insertBefore, replaceChild operations; 
    *  arguments - action-name, the node's parent, the node along with the
    *              reference node as the argumnts.
    */
   
   public Node Wrapper(String action, Node parent, Node newnode, Node ref) {
     /* xpath     - stores the Xpath expression for the parent-node, target node
      * argaction - a formal name corresponding to the action specified.
      */ 
     String xpath[], argaction;
     //boolean flag = false; 

     /* returnNode - stores the Diff-DOM node returned after processing.
      * diffNode   - the node that is stored in the DOM (Diff-DOM) after
      *              being processed.
      */              
     Node returnNode = null, diffNode;
     Element diffRoot;
    
     /* The following is the general sequence of operations involved for any action.
      * 1) First remove the event-listener.
      * 2) Perform the actual action.
      * 3) Add the event-listener back.
      */
     
     removeDiffListener(); 
     xpath = new String[2];

      /* diff - creates a Document object for every newly created node of
       *        the Diff-DOM, with the root nodes as "diff".
       */

     DiffDOMDoc diff = new DiffDOMDoc();
     diffDoc = diff.getNewDiffDoc();

     //diffChild - the individual node that occurs in the DOM modelling the "diff".
     DiffDomNode  diffChild = new DiffDomNode(diffDoc, newnode); 
     diffRoot = (Element)diffDoc.getFirstChild();
     System.arraycopy(diffChild.getXPath(ref, newnode, false), 0, xpath, 0, 2);
     
     if ( action.equals("insertBefore") ) {
        argaction = DiffNodeInfo.ACTION_INSERTBEFORE;  
     }
     else {
        argaction = DiffNodeInfo.ACTION_REPLACE;
     }

     //nodeInfo - stores the context information gathered above.
     DiffNodeInfo nodeInfo = new DiffNodeInfo(argaction, ref, newnode, null, null, null, null);
     //key = xpath[1]; 
     
     if (newnode.getNodeType() == Node.ELEMENT_NODE) {
        /* Key exists for an element that has been recently created.
           For an already created node (from an existing XML Document), 
	   xpath[1] MAY NOT exist.    
        */
        if (  !table.containsKey(xpath[1]) ) {
          diffNode = diffChild.wrapper(nodeInfo);
	  
          HashValue  hashEntry = new HashValue();
          //Store this "diff" node as the value of the hashtable;
          hashEntry.domNodeRef = diffNode;
          table.put(xpath[1], hashEntry);

          /* If the target-node is appended to a different subtree all 
	   * its descendants must have the latest xpath attribs. 
	   */
           if (newnode.hasChildNodes()) {
              System.out.println("to updateXPAth..");
              diffChild.updateXPathAttribs(newnode, table); 
          } //if(3) 
        } //if(2)
        else {
             /* If the key already exists:
                1) If action=node-replace; updateXPathAttributes() of the 
		   child-nodes of this node; remove the keys, childnodes of the 
		   old-node ;
             */ 
             Node storedNode = ((HashValue)table.get(xpath[1])).domNodeRef;
             nodeInfo.storedNode = storedNode;
             diffChild.setDiffNode(storedNode);
             diffNode = diffChild.wrapper(nodeInfo);

             //Point to the Document node of this Diff-DOM node.
             diff.diffDoc = storedNode.getOwnerDocument();
             if (diffNode == null) {
               /*  1)   Remove all the keys, childNodes associated with the old node and 
                        the  children.
                   2)   Update the XPath attribs of the new node.
               */
                if ( argaction.equals(DiffNodeInfo.ACTION_REPLACE) ) {
                  diffChild.updateXPathAttribs(newnode, table);
                  diffChild.deleteDiffChildren(ref, table);
                  returnNode = parent.replaceChild(newnode, ref);
                }
                else if ( argaction.equals(DiffNodeInfo.ACTION_INSERTBEFORE) ) { 
                  /* If the target-node is a subtree (OR) a single node that is 
		   * inserted-before a node, just update its "ref" Xpath attrib. 
                   */
                   diffChild.updateRefXPathAttrib(newnode,ref, table);
                   returnNode = parent.insertBefore(newnode, ref); 
                }
              }  
        }
     } //ELEMENT_NODE 
    addDiffListener();
    return (returnNode);
  } //Wrapper.

}// XMLDiff
